#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <queue.h>

/**
 * @file queue.c
 * @brief File di implementazione dell'interfaccia per la coda
 */



/* ------------------- funzioni di utilita' -------------------- */

static Node_t *allocNode()                   { return malloc(sizeof(Node_t));  }
static Queue_t *allocQueue()                 { return malloc(sizeof(Queue_t)); }
static void freeNode(Node_t *node)           { free((void*)node); }
static void LockQueue(Queue_t *q)            { pthread_mutex_lock(&q->qlock);   }
static void UnlockQueue(Queue_t *q)          { pthread_mutex_unlock(&q->qlock); }
static void UnlockQueueAndWait(Queue_t *q)   { pthread_cond_wait(&q->qcond, &q->qlock); }
static void UnlockQueueAndSignal(Queue_t *q) {
    pthread_cond_signal(&q->qcond);
    pthread_mutex_unlock(&q->qlock);
}

/* ------------------- interfaccia della coda ------------------ */

Queue_t *initQueue() {
    Queue_t *q = allocQueue();
    if (!q) return NULL;
    q->head = allocNode();
    if (!q->head) return NULL;
    q->head->data = NULL; 
    q->head->next = NULL;
    q->tail = q->head;    
    q->qlen = 0;
    if (pthread_mutex_init(&q->qlock, NULL) != 0) {
	perror("mutex init");
	return NULL;
    }
    if (pthread_cond_init(&q->qcond, NULL) != 0) {
	perror("mutex cond");
	return NULL;
    }    
    return q;
}

void deleteQueue(Queue_t *q) {
    while(q->head != q->tail) {
	Node_t *p = (Node_t*)q->head;
	q->head = q->head->next;
	freeNode(p);
    }
    if (q->head) freeNode((void*)q->head);
    pthread_mutex_destroy(&q->qlock);
    pthread_cond_destroy(&q->qcond);
    free(q);
}

int push(Queue_t *q, void *data) {
    Node_t *n = allocNode();
    n->data = data; 
    n->next = NULL;
    LockQueue(q);
    q->tail->next = n;
    q->tail       = n;
    q->qlen      += 1;
    UnlockQueueAndSignal(q);
    return 0;
}

void *pop(Queue_t *q) {        
    LockQueue(q);
    while(q->head == q->tail) {
	UnlockQueueAndWait(q);
    }
    // locked
    assert(q->head->next);
    Node_t *n  = (Node_t *)q->head;
    void *data = (q->head->next)->data;
    q->head    = q->head->next;
    q->qlen   -= 1;
    assert(q->qlen>=0);
    UnlockQueue(q);
    freeNode(n);
    return data;
} 

// WARNING: accesso in sola lettura non in mutua esclusione
unsigned long length(Queue_t *q) {
    unsigned long len = q->qlen;
    return len;
}

